import numpy as np
from sklearn.metrics import classification_report
from sklearn.isotonic import IsotonicRegression
from .VennABERS import ScoresToMultiProbs
from .calibrutils import reliabubble, calibration_errors
from scipy.special import softmax


class IVAP:

	def __init__(self, y_cal, y_test, y_cal_pred, y_test_pred, label=1, apply_softmax=False):
		self.y_cal = y_cal
		self.y_test = y_test
		self.y_cal_pred = y_cal_pred
		self.y_test_pred = y_test_pred
		self.label = label
		if apply_softmax == True:
			self.y_cal_pred = softmax(self.y_cal_pred, axis=1)
			self.y_test_pred = softmax(self.y_test_pred, axis=1)
		self.y_cal_scores = self.y_cal_pred[:, self.label]
		self.y_test_scores = self.y_test_pred[:, self.label]
		self.p_0, self.p_1 = self.__get_multiprobs()
		self.p_single = self.__compute_single_prob()


	def __get_multiprobs(self):
		calibr_points = list(zip(self.y_cal_scores, self.y_cal))
		return ScoresToMultiProbs(calibr_points, self.y_test_scores)

	def __compute_single_prob(self):
		return self.p_1 / (1 - self.p_0 + self.p_1)

	def prob_uncertainty(self):
		return np.abs(self.p_0 - self.p_1)

	def print_performance(self, digits=2):
		y_hat_test = np.round(self.p_single).astype(int)
		print(classification_report(self.y_test, y_hat_test, digits=digits))

	def compute_calibration_errors(self, num_bins=10):
		return calibration_errors(self.p_single, self.y_test, num_bins)

	def plot_reliabubble(self, num_bins=10, size_max=20, font_size=12, title=None):
		return reliabubble(self.p_single, self.y_test, num_bins, size_max, font_size, title=title)

	def compare_with_isotonic_regression(self, num_bins=10):
		iso_reg = IsotonicRegression(out_of_bounds='clip').fit(self.y_cal_scores, self.y_cal)
		y_test_iso = iso_reg.predict(self.y_test_scores)
		print('Comparison with direct isotonic regression:')
		print(calibration_errors(y_test_iso, self.y_test, num_bins, include_log_loss=False))
		return reliabubble(y_test_iso, self.y_test, num_bins, title='Isotonic Regression')


			


import numpy as np
import plotly.graph_objects as go
import plotly.express as px
from sklearn.metrics import log_loss, brier_score_loss


def reliabubble(scores, y_test, num_bins=10, size_max=20, font_size=12, title=None):
	# Create 10 bins (n. 1 to n. 10) such that 0.0 and 1.0 values are included
	# Bin n. 0 would get values < 0.0 (it shouldn't happen).
	bins = np.linspace(0, 1 + 1e-8, num=num_bins + 1)  
	whichbin = np.digitize(scores, bins)  # [bin of example forall examples]
	counts = np.bincount(whichbin, minlength=len(bins))
	proportions = counts / len(y_test)

	true_per_bin = np.bincount(whichbin, weights=y_test, minlength=len(bins))

	accuracy = np.zeros(len(bins))  # Initialise to use the "where" condition in np.divide()
	accuracy = np.divide(true_per_bin, counts, where=counts!=0)

	score_sum_per_bin = np.bincount(whichbin, weights=scores, minlength=len(bins))

	confidence = np.zeros(len(bins))
	confidence = np.divide(score_sum_per_bin, counts, where=counts!=0)

    # 2 equally large bubbles on 2 different plots may reflect different numbers
    # this is due to plotly using relative frequencies to determine the bubble size (understandably)
    # how can we link two graphs' relative  frequencies so that it's easy to compare them? facets?
	fig = px.scatter(x=confidence, y=accuracy, size=counts, width=700, height=500,
		labels={
			'x':'probability',
			'y':'accuracy'
		},
		title = title,
		size_max = size_max)
	fig.update_layout(font_size=font_size)
	fig.add_trace(
		go.Scatter(x=bins,
			y=bins,
			opacity=0.2,
			line=dict(
				color='gray',
				dash='dash'
				),
			showlegend=False
			)
		)
	return fig


def calibration_errors(scores, y_test, num_bins=10, include_log_loss=True):
	errors = {}
	bins = np.linspace(0, 1 + 1e-8, num=num_bins + 1)
	whichbin = np.digitize(scores, bins)  # [bin of example forall examples]
	counts = np.bincount(whichbin, minlength=len(bins))
	assert counts[0] == 0, 'There seems to be a score < 0.0 somewhere.'
	proportions = counts / len(y_test)
	nonzero = counts != 0

	true_per_bin = np.bincount(whichbin, weights=y_test, minlength=len(bins))
	accuracy = np.divide(true_per_bin[nonzero], counts[nonzero])

	score_sum_per_bin = np.bincount(whichbin, weights=scores, minlength=len(bins))
	confidence = np.divide(score_sum_per_bin[nonzero], counts[nonzero])  # mean confidence per bin

	ece = np.sum(np.multiply(proportions[nonzero], np.abs(confidence - accuracy)))
	mce = np.max(np.abs(confidence - accuracy))

	errors['ECE'] = ece
	errors['MCE'] = mce
	if include_log_loss == True:
		errors['log loss'] = log_loss(y_test, scores)
	errors['Brier loss'] = brier_score_loss(y_test, scores)

	return errors




# Straight-forward implementation of IVAP algorithm described in:
# Large-scale probabilistic prediction with and without validity guarantees, Vovk et al.
# https://arxiv.org/pdf/1511.00213.pdf
#
# Paolo Toccaceli
#
# https://github.com/ptocca/VennABERS
#
# 2020-07-09: Fixed bug in p0 calculation

import numpy as np

# Some elementary functions to speak the same language as the paper
# (at some point we'll just replace the occurrence of the calls with the function body itself)
def push(x,stack):
    stack.append(x)

    
def pop(stack):
    return stack.pop()


def top(stack):
    return stack[-1]


def nextToTop(stack):
    return stack[-2]


# perhaps inefficient but clear implementation
def nonleftTurn(a,b,c):   
    d1 = b-a
    d2 = c-b
    return np.cross(d1,d2)<=0


def nonrightTurn(a,b,c):   
    d1 = b-a
    d2 = c-b
    return np.cross(d1,d2)>=0


def slope(a,b):
    ax,ay = a
    bx,by = b
    return (by-ay)/(bx-ax)


def notBelow(t,p1,p2):
    p1x,p1y = p1
    p2x,p2y = p2
    tx,ty = t
    m = (p2y-p1y)/(p2x-p1x)
    b = (p2x*p1y - p1x*p2y)/(p2x-p1x)
    return (ty >= tx*m+b)

kPrime = None

# Because we cannot have negative indices in Python (they have another meaning), I use a dictionary

def algorithm1(P):
    global kPrime
    
    S = []
    P[-1] = np.array((-1,-1))
    push(P[-1],S)
    push(P[0],S)
    for i in range(1,kPrime+1):
        while len(S)>1 and nonleftTurn(nextToTop(S),top(S),P[i]):
            pop(S)
        push(P[i],S)
    return S


def algorithm2(P,S):
    global kPrime
    
    Sprime = S[::-1]     # reverse the stack

    F1 = np.zeros((kPrime+1,))
    for i in range(1,kPrime+1):
        F1[i] = slope(top(Sprime),nextToTop(Sprime))
        P[i-1] = P[i-2]+P[i]-P[i-1]
        if notBelow(P[i-1],top(Sprime),nextToTop(Sprime)):
            continue
        pop(Sprime)
        while len(Sprime)>1 and nonleftTurn(P[i-1],top(Sprime),nextToTop(Sprime)):
            pop(Sprime)
        push(P[i-1],Sprime)
    return F1


def algorithm3(P):
    global kPrime

    S = []
    push(P[kPrime+1],S)
    push(P[kPrime],S)
    for i in range(kPrime-1,0-1,-1):  # k'-1,k'-2,...,0
        while len(S)>1 and nonrightTurn(nextToTop(S),top(S),P[i]):
            pop(S)
        push(P[i],S)
    return S


def algorithm4(P,S):
    global kPrime
    
    Sprime = S[::-1]     # reverse the stack
    
    F0 = np.zeros((kPrime+1,))
    for i in range(kPrime,1-1,-1):   # k',k'-1,...,1
        F0[i] = slope(top(Sprime),nextToTop(Sprime))
        P[i] = P[i-1]+P[i+1]-P[i]
        if notBelow(P[i],top(Sprime),nextToTop(Sprime)):
            continue
        pop(Sprime)
        while len(Sprime)>1 and nonrightTurn(P[i],top(Sprime),nextToTop(Sprime)):
            pop(Sprime)
        push(P[i],Sprime)
    return F0


def prepareData(calibrPoints):
    global kPrime
    
    ptsSorted = sorted(calibrPoints)
    
    xs = np.fromiter((p[0] for p in ptsSorted),float)
    ys = np.fromiter((p[1] for p in ptsSorted),float)
    ptsUnique,ptsIndex,ptsInverse,ptsCounts = np.unique(xs, 
                                                        return_index=True,
                                                        return_counts=True,
                                                        return_inverse=True)
    a = np.zeros(ptsUnique.shape)
    np.add.at(a,ptsInverse,ys)
    # now a contains the sums of ys for each unique value of the objects
    
    w = ptsCounts
    yPrime = a/w
    yCsd = np.cumsum(w*yPrime)   # Might as well do just np.cumsum(a)
    xPrime = np.cumsum(w)
    kPrime = len(xPrime)
    
    return yPrime,yCsd,xPrime,ptsUnique


def computeF(xPrime,yCsd):
    global kPrime
    P = {0:np.array((0,0))}
    P.update({i+1:np.array((k,v)) for i,(k,v) in enumerate(zip(xPrime,yCsd))})
    
    S = algorithm1(P)
    F1 = algorithm2(P,S)
    
    P = {0:np.array((0,0))}
    P.update({i+1:np.array((k,v)) for i,(k,v) in enumerate(zip(xPrime,yCsd))})    
    P[kPrime+1] = P[kPrime] + np.array((1.0,0.0))    # The paper says (1,1)
    
    S = algorithm3(P)
    F0 = algorithm4(P,S)
    
    return F0,F1


def getFVal(F0,F1,ptsUnique,testObjects):
    pos0 = np.searchsorted(ptsUnique,testObjects,side='left')
    pos1 = np.searchsorted(ptsUnique[:-1],testObjects,side='right')+1
    return F0[pos0],F1[pos1]


def ScoresToMultiProbs(calibrPoints,testObjects):
    # sort the points, transform into unique objects, with weights and updated values
    yPrime,yCsd,xPrime,ptsUnique = prepareData(calibrPoints)
    
    # compute the F0 and F1 functions from the CSD
    F0,F1 = computeF(xPrime,yCsd)
    
    # compute the values for the given test objects
    p0,p1 = getFVal(F0,F1,ptsUnique,testObjects)
                    
    return p0,p1


import numpy as np
from sklearn.metrics import classification_report
import plotly.express as px
import plotly.graph_objects as go
# import matplotlib.pyplot as plt
# import seaborn as sns
import pandas as pd


class ConformalPredictor():

	def __init__(self, y_cal_proba, y_cal, y_test_proba, y_test, smoothed=False, mondrian=True):
		self.y_cal_proba = y_cal_proba
		self.y_cal = y_cal
		self.y_test_proba = y_test_proba
		self.y_test = y_test
		self.labels = [i for i in range(y_cal_proba.shape[1])]
		self.alphas = np.zeros(len(y_cal))  # nonconformity scores for calibration set
		self.ranks = np.zeros((len(y_test), len(self.labels)))
		self.p_values = np.zeros((len(y_test), len(self.labels)))
		self.smoothed = smoothed
		self.mondrian = mondrian

	def fit(self, ncf_measure):
		if self.mondrian:
			for l in self.labels:
				ncm_cal = ncf_measure(self.y_cal_proba, self.y_cal)
				ncm_cal = ncm_cal[self.y_cal == l]
				self.alphas = ncm_cal
				array_of_trues = np.full(len(self.y_test), l)  # Trick: run ncf_measure over *all* the test examples
				ncm_test = ncf_measure(self.y_test_proba, array_of_trues)
				# sorted_alpha_cal = np.sort(ncm_cal)
				# ranks = len(ncm_cal) - np.searchsorted(sorted_alpha_cal, ncm_test) + 1
				ranks = self.compute_ranks(ncm_test)
				self.ranks[:, int(l)] = ranks
				p_value = ranks / (len(ncm_cal) + 1)
				self.p_values[:, int(l)] = p_value
		else:
			ncm_cal = ncf_measure(self.y_cal_proba, self.y_cal)
			self.alphas = ncm_cal
			for l in self.labels:
				array_of_trues = np.full(len(self.y_test), l)
				ncm_test = ncf_measure(self.y_test_proba, array_of_trues)
				# sorted_alpha_cal = np.sort(ncm_cal)
				# ranks = len(ncm_cal) - np.searchsorted(sorted_alpha_cal, ncm_test) + 1
				ranks = self.compute_ranks(ncm_test)
				self.ranks[:, int(l)] = ranks
				p_value = ranks / (len(ncm_cal) + 1)
				self.p_values[:, int(l)] = p_value

	def compute_ranks(self, ncm_test):
		sorted_alpha_cal = np.sort(self.alphas)
		greater_alphas = np.searchsorted(sorted_alpha_cal, ncm_test, side='right')
		geq_alphas = np.searchsorted(sorted_alpha_cal, ncm_test, side='left')
		if self.smoothed:
			ties = geq_alphas - greater_alphas
			random_ties = ties * np.random.uniform(0, 1, len(ties))  # Element by element

			ranks = len(self.alphas) + 1 - greater_alphas - random_ties
		else:
			ranks = len(self.alphas) + 1 - geq_alphas
		return ranks

	def predict(self, epsilon):
		return self.p_values >= epsilon

	def point_prediction_performance(self, digits=2):
		y_hat_test = np.argmax(self.p_values, axis=1)
		print(classification_report(self.y_test, y_hat_test, digits=digits))

	def s_criterion(self):	
		return np.average(self.p_values)

	# TODO URGENT: Check when number of labels > 2
	# The average false p-value: the average of the p-values for *ALL*
	# postulated labels in the test set except for the true labels.
	def of_criterion(self):
		y_test_false = np.logical_not(self.y_test).astype(int)
		false_p_values = np.take_along_axis(self.p_values, y_test_false[:, None], axis=1)
		return np.average(false_p_values)

	def n_criterion(self):
		sig_levels = []
		avg_sizes = []
		for i in range(0, 101, 1):
			eps = i/100
			sig_levels.append(eps)
			regions = self.p_values > eps
			region_size = np.sum(regions, axis=1)
			avg_sizes.append(np.sum(region_size) / len(self.y_test))
		return pd.DataFrame({'significance':sig_levels, 'prediction size':avg_sizes}) 

	def top_high_confidence(self, n=10):
		second_largest = np.sort(self.p_values, axis=1)[:, -2]
		return np.argsort(second_largest)[:n]  # it's in increasing order

	def top_low_confidence(self, n=10):
		second_largest = np.sort(self.p_values, axis=1)[:, -2]
		return np.argsort(second_largest)[-n:]  # it's in increasing order

	def top_high_credibility(self, n=10):
		return np.argsort(np.max(self.p_values, axis=1))[-n:]

	def top_low_credibility(self, n=10):
		return np.argsort(np.max(self.p_values, axis=1))[:n]

	def plot_validity(self, labelwise=False):  # TODO: granularity;
		significance = []
		error_rates = []
		pd_labels = []

		if labelwise:
			labels = self.labels
		else:
			labels = ['all']
		
		for label in labels:	
			for i in range(0, 101, 2):
				eps = i/100
				significance.append(eps)
				if labelwise:
					regions = self.p_values[self.y_test==label, :] > eps
					true_label_not_present = np.invert(regions[np.arange(len(regions)), self.y_test[self.y_test==label]]) 
					error_rate = np.sum(true_label_not_present) / len(self.y_test[self.y_test==label])
				else:
					regions = self.p_values > eps
					true_label_not_present = np.invert(regions[np.arange(len(regions)), self.y_test]) 
					error_rate = np.sum(true_label_not_present) / len(self.y_test)
				error_rates.append(error_rate)
				pd_labels.append(str(label))

		df = pd.DataFrame({'significance':significance, 'error rate':error_rates, 'label':pd_labels})

		fig = px.scatter(df, x='significance', y='error rate', color='label', width=500, height=400,  # 700, 500
			template='none',
			color_discrete_sequence=px.colors.qualitative.T10
		)
		fig.add_trace(
			go.Scatter(x=significance,
				y=significance,
				opacity=0.2,
				line=dict(
					color='gray',
					dash='dash'
					),
				showlegend=False
				)
		)
		fig.update_layout(
			legend=dict(
			    yanchor="top",
			    y=0.94,
			    xanchor="left",
			    x=0.1),
			# font_family='Officina Sans ITC Book'
		)
		fig.update_xaxes(nticks=10)

		# fig = sns.scatterplot(data=df, x='significance', y='error rate', hue='label')

		return fig




	# def old_plot_validity(self):
	# 	error_rates = []
	# 	significance = []
	# 	sizes = []
	# 	empties = []
	# 	for i in range(0, 101, 2):
	# 		eps = i/100
	# 		significance.append(eps)
	# 		regions = self.p_values > eps
	# 		no_true_included = np.invert(np.take_along_axis(regions, self.y_test[:, None], axis=1))
	# 		error_rate = np.sum(no_true_included) / len(self.y_test)
	# 		error_rates.append(error_rate)
	# 		region_size = np.sum(regions, axis=1)
	# 		sizes.append(np.sum(region_size) / len(self.y_test))
	# 		empties.append(np.sum(region_size == 0) / len(self.y_test))

	# 	fig, axs = plt.subplots(3, sharex=True, figsize=(5,10))

	# 	axs[0].plot(significance, significance, color='lightgray', linestyle='dashed')
	# 	axs[0].plot(significance, error_rates)
	# 	# ax[0].xlabel('Confidence')
	# 	# ax[0].ylabel('Error rate')
	# 	axs[0].legend(['Theoretical', 'Actual error rate'])
	# 	# ax[0].rcParams.update({'font.size': 12})

	# 	axs[1].plot(significance, sizes)
	# 	axs[1].legend(['Average prediction size'])
	# 	axs[2].plot(significance, empties)
	# 	axs[2].legend(['Empty predictions'])

	# 	for ax in axs.flat:
	# 		ax.grid(b=True, color='#999999', linestyle='-', alpha=0.2)

	# 	axs[2].set(xlabel='significance level')





import numpy as np
from scipy.special import softmax

def negative_logit(y_proba, y):
    return -y_proba[np.arange(len(y_proba)), y]

def negative_softmax(y_proba, y):
    y_proba = softmax(y_proba, axis=1)
    return -y_proba[np.arange(len(y_proba)), y]

def logit_ratio(y_proba, y):
    idx_of_true = np.zeros(y_proba.shape, dtype=bool)
    idx_of_true[np.arange(len(y_proba)), y] = True
    num = y_proba[np.invert(idx_of_true)]
    if y_proba.shape[1] > 2:
        num = np.sum(num, axis=1)
    den = y_proba[idx_of_true] + 1e6
    return np.divide(num, den)

def margin_error_func(y_proba, y):  # y is:  i) true y_cal  and  ii) y_test candidate
    prob_true = y_proba[np.arange(len(y_proba)), y]
    # print(prob_true)
    y_probba = y_proba.copy()
    np.put_along_axis(y_probba, y[:, None], -np.inf, axis=1)  # "exclude" probs of true labels
    # print(f'y_probba now: \n{y_probba}')
    max_among_false = np.max(y_probba, axis=1)
    # print(f'Max among false: \n{max_among_false}')
    return 0.5 - ((prob_true - max_among_false) / 2)

# score is the max difference between true label and softmax prediction
def l_infinity(y_proba, y):
    pass




import numpy as np
from icp import ConformalPredictor
from sklearn.metrics import classification_report, matthews_corrcoef, accuracy_score, precision_recall_fscore_support, f1_score
from sklearn.model_selection import KFold
from transformers import AutoTokenizer, AutoModelForSequenceClassification, TrainingArguments, Trainer
import plotly.express as px
import plotly.graph_objects as go
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

# ideally:
class CrossConformalPredictor():
	def __init__(self, X_train, y_train, X_val, y_val, X_test, y_test, train_func, pred_func, n_folds, mondrian=False):
		pass

class CrossConformalPredictor():

	def __init__(self, X_train, y_train, X_test, y_test, validation_set, model_name, tokenizer, labels, n_folds=3, mondrian=True):
		self.X_train = X_train
		self.y_train = y_train
		self.X_test = X_test
		self.y_test = y_test
		self.validation_set = validation_set
		self.model_name = model_name
		self.tokenizer = tokenizer
		self.labels = labels
		self.num_labels = len(self.labels)
		self.p_values = np.zeros((len(y_test), len(labels)))
		self.n_folds = n_folds
		self.mondrian = mondrian
		self.proper_trains = DatasetDict()
		self.calis = DatasetDict()
		self.preds = {}
		kf = KFold(n_splits=self.n_folds, shuffle=True, random_state=1986)
		for f, (pt, cal) in enumerate(kf.split(data['train'])):
			    self.proper_trains[f] = X_train.select(pt)
			    self.calis[f] = X_train.select(cal)



	def train(self, train_args):
		for i in range(self.n_folds):
		    preds[i] = {}

		    model = AutoModelForSequenceClassification.from_pretrained(self.model_name, num_labels=self.num_labels)
		    model.to(device)

		    trainer = Trainer(
		        model,
		        args,
		        train_dataset=proper_trains[i],
		        eval_dataset=self.validation_set,
		        tokenizer=self.tokenizer,
		        compute_metrics=compute_metrics
		    )
		    trainer.train()
		    
		    # conformal prediction
		    y_cal = calis[i]['label'].numpy()
		    y_test = data['test']['label'].numpy()
		    cal_predictions = trainer.predict(calis[i])
		    y_cal_proba = cal_predictions.predictions
		    y_test_proba = trainer.predict(data['test']).predictions
		    labels = [i for i in range(self.num_labels)]
		    
		    preds[i]['y_cal'] = y_cal
		    preds[i]['y_cal_proba'] = y_cal_proba
		    preds[i]['y_test'] = y_test
		    preds[i]['y_test_proba'] = y_test_proba


	def compute_metrics(eval_pred):
	    predictions, labels = eval_pred
	    preds = np.argmax(predictions, axis=1)
	    precision, recall, f1_macro, _ = precision_recall_fscore_support(labels, preds, average='macro')
	    acc = accuracy_score(labels, preds)
	    f1_weighted = f1_score(labels, preds, average='weighted')
	    return {
	        'accuracy': acc,
	        'f1_macro': f1_macro,
	        'f1_weighted': f1_weighted,
	        'precision': precision,
	        'recall': recall
	    }



	def fit(self, ncf_measure):
		for i in range(n_folds):
		    preds[i]['icp'] = ConformalPredictor(preds[i]['y_cal_proba'], preds[i]['y_cal'], 
		                                      preds[i]['y_test_proba'], preds[i]['y_test'], 
		                                      self.labels, mondrian=self.mondrian)
		    preds[i]['icp'].fit(ncf_measure)
		    preds[i]['ranks'] = preds[i]['icp'].ranks - 1  # necessary for cross-conformal

		    rz = np.sum(np.array([preds[i]['ranks'] for i in range(self.n_folds)]), axis=0)  # sum ranks over all folds

		    if is_mondrian:
			    for label in labels:
			        self.p_values[:, label] = (rz[:, label] + 1) / (int(sum(X_train['label'] == label)) + 1)
			else:
			    self.p_values = (rz + 1) / (len(self.X_train) + 1)


	def point_prediction_performance(self, digits=2):
		y_hat_test = np.argmax(self.p_values, axis=1)
		print(classification_report(self.y_test, y_hat_test, digits=digits))

	def average_false_pvalue(self):
		y_test_false = np.logical_not(self.y_test).astype(int)
		false_p_values = np.take_along_axis(self.p_values, y_test_false[:, None], axis=1)
		return np.average(false_p_values)

	def top_high_confidence(self, n=10):
		second_largest = np.sort(self.p_values, axis=1)[:, -2]
		return np.argsort(second_largest)[:n]  # it's in increasing order

	def top_low_confidence(self, n=10):
		second_largest = np.sort(self.p_values, axis=1)[:, -2]
		return np.argsort(second_largest)[-n:]  # it's in increasing order

	def top_high_credibility(self, n=10):
		return np.argsort(np.max(self.p_values, axis=1))[-n:]

	def top_low_credibility(self, n=10):
		return np.argsort(np.max(self.p_values, axis=1))[:n]


	def plot_validity(self, labelwise=False):  # TODO: granularity;
		significance = []
		error_rates = []
		pd_labels = []

		if labelwise:
			labels = self.labels
		else:
			labels = ['all']
		
		for label in labels:	
			for i in range(0, 101, 2):
				eps = i/100
				significance.append(eps)
				if labelwise:
					regions = self.p_values[self.y_test==label, :] > eps
					true_label_not_present = np.invert(regions[np.arange(len(regions)), self.y_test[self.y_test==label]]) 
					error_rate = np.sum(true_label_not_present) / len(self.y_test[self.y_test==label])
				else:
					regions = self.p_values > eps
					true_label_not_present = np.invert(regions[np.arange(len(regions)), self.y_test]) 
					error_rate = np.sum(true_label_not_present) / len(self.y_test)
				error_rates.append(error_rate)
				pd_labels.append(str(label))

		df = pd.DataFrame({'significance':significance, 'error rate':error_rates, 'label':pd_labels})

		fig = px.scatter(df, x='significance', y='error rate', color='label', width=500, height=400,  # 700, 500
			template='none',
			color_discrete_sequence=px.colors.qualitative.T10
		)
		fig.add_trace(
			go.Scatter(x=significance,
				y=significance,
				opacity=0.2,
				line=dict(
					color='gray',
					dash='dash'
					),
				showlegend=False
				)
		)
		fig.update_layout(
			legend=dict(
			    yanchor="top",
			    y=0.94,
			    xanchor="left",
			    x=0.1),
			# font_family='Officina Sans ITC Book'
		)
		fig.update_xaxes(nticks=10)

		# fig = sns.scatterplot(data=df, x='significance', y='error rate', hue='label')

		return fig
